// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    expr::{ops::ArrayIndex, Expr, ToExpr},
    int::{DynSize, KnownSize, Size},
    intern::{Intern, Interned},
    module::{ModuleBuilder, NormalModule},
    source_location::SourceLocation,
    ty::{
        CanonicalType, MatchVariantWithoutScope, StaticType, Type, TypeProperties, TypeWithDeref,
    },
    util::ConstUsize,
};
use std::ops::Index;

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
struct ArrayImpl<T: Type, Len: Size> {
    element: T,
    len: Len::SizeType,
    type_properties: TypeProperties,
}

#[derive(Copy, Clone, PartialEq, Eq, Hash)]
pub struct ArrayType<T: Type = CanonicalType, Len: Size = DynSize> {
    impl_: Interned<ArrayImpl<T, Len>>,
}

impl<T: Type, Len: Size> std::fmt::Debug for ArrayType<T, Len> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let ArrayImpl {
            element, len: _, ..
        } = *self.impl_;
        write!(f, "Array<{element:?}, {}>", self.len())
    }
}

pub type Array<
    T = CanonicalType,
    const LEN: usize = { <DynSize as crate::util::GenericConstUsize>::VALUE },
> = ArrayType<T, ConstUsize<LEN>>;

#[allow(non_upper_case_globals)]
pub const Array: ArrayWithoutGenerics = ArrayWithoutGenerics;

impl<T: Type, Len: Size> ArrayType<T, Len> {
    pub fn new(element: T, len: Len::SizeType) -> Self {
        let element_props = element.canonical().type_properties();
        let type_properties = TypeProperties {
            is_passive: element_props.is_passive,
            is_storable: element_props.is_storable,
            is_castable_from_bits: element_props.is_castable_from_bits,
            bit_width: element_props
                .bit_width
                .checked_mul(Len::as_usize(len))
                .expect("array too big"),
        };
        Self {
            impl_: ArrayImpl {
                element,
                len,
                type_properties,
            }
            .intern_sized(),
        }
    }
    pub fn element(&self) -> &T {
        &self.impl_.element
    }
    pub fn len(self) -> usize {
        Len::as_usize(self.impl_.len)
    }
    pub fn type_properties(self) -> TypeProperties {
        self.impl_.type_properties
    }
    pub fn as_dyn_array(self) -> Array {
        Array::new_dyn(self.element().canonical(), self.len())
    }
}

impl<T: Type, Len: KnownSize> ArrayType<T, Len> {
    pub fn new_static(element: T) -> Self {
        Self::new(element, Len::SizeType::default())
    }
}

impl<T: StaticType, Len: KnownSize> StaticType for ArrayType<T, Len> {
    fn static_type() -> Self {
        Self::new_static(T::static_type())
    }
}

impl<T: Type> Array<T> {
    pub fn new_dyn(element: T, len: usize) -> Self {
        Self::new(element, len)
    }
}

impl<T: Type, Len: Size> Type for ArrayType<T, Len> {
    type MaskType = ArrayType<T::MaskType, Len>;
    type MatchVariant = Len::ArrayMatch<T>;
    type MatchActiveScope = ();
    type MatchVariantAndInactiveScope = MatchVariantWithoutScope<Len::ArrayMatch<T>>;
    type MatchVariantsIter = std::iter::Once<Self::MatchVariantAndInactiveScope>;

    fn match_variants(
        this: Expr<Self>,
        module_builder: &mut ModuleBuilder<NormalModule>,
        source_location: SourceLocation,
    ) -> Self::MatchVariantsIter {
        let base = Expr::as_dyn_array(this);
        let base_ty = Expr::ty(base);
        let _ = module_builder;
        let _ = source_location;
        let retval = Vec::from_iter(
            (0..base_ty.len()).map(|i| ArrayIndex::new_unchecked(base, i).to_expr()),
        );
        std::iter::once(MatchVariantWithoutScope(
            Len::ArrayMatch::<T>::try_from(retval)
                .ok()
                .expect("unreachable"),
        ))
    }

    fn mask_type(&self) -> Self::MaskType {
        ArrayType::new(self.element().mask_type(), self.impl_.len)
    }

    fn canonical(&self) -> CanonicalType {
        CanonicalType::Array(Array::new_dyn(self.element().canonical(), self.len()))
    }

    #[track_caller]
    fn from_canonical(canonical_type: CanonicalType) -> Self {
        let CanonicalType::Array(array) = canonical_type else {
            panic!("expected array");
        };
        Self::new(
            T::from_canonical(*array.element()),
            Len::from_usize(array.len()),
        )
    }
    fn source_location() -> SourceLocation {
        SourceLocation::builtin()
    }
}

impl<T: Type, Len: Size> TypeWithDeref for ArrayType<T, Len> {
    fn expr_deref(this: &Expr<Self>) -> &Self::MatchVariant {
        let base = Expr::as_dyn_array(*this);
        let base_ty = Expr::ty(base);
        let retval = Vec::from_iter(
            (0..base_ty.len()).map(|i| ArrayIndex::new_unchecked(base, i).to_expr()),
        );
        Interned::<_>::into_inner(Intern::intern_sized(
            Len::ArrayMatch::<T>::try_from(retval)
                .ok()
                .expect("unreachable"),
        ))
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug, Default)]
pub struct ArrayWithoutGenerics;

impl<T: Type> Index<T> for ArrayWithoutGenerics {
    type Output = ArrayWithoutLen<T>;

    fn index(&self, element: T) -> &Self::Output {
        Interned::<_>::into_inner(Intern::intern_sized(ArrayWithoutLen { element }))
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct ArrayWithoutLen<T: Type> {
    element: T,
}

impl<T: Type> Index<usize> for ArrayWithoutLen<T> {
    type Output = Array<T>;

    fn index(&self, len: usize) -> &Self::Output {
        Interned::<_>::into_inner(Intern::intern_sized(Array::new_dyn(self.element, len)))
    }
}
