// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    array::{Array, ArrayType},
    int::{Bool, IntType, SIntType, SIntValue, Size, UIntType, UIntValue},
    intern::{Intern, Interned},
    ty::{CanonicalType, Type},
};

pub mod ops;

macro_rules! expr_enum {
    (
        pub enum $ExprEnum:ident {
            $($Variant:ident($VariantTy:ty),)*
        }
    ) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub enum $ExprEnum {
            $($Variant($VariantTy),)*
        }

        $(impl From<$VariantTy> for $ExprEnum {
            fn from(v: $VariantTy) -> Self {
                Self::$Variant(v)
            }
        })*
    };
}

expr_enum! {
    pub enum ExprEnum {
        UIntLiteral(Interned<UIntValue>),
        SIntLiteral(Interned<SIntValue>),
        BoolLiteral(bool),
        BundleLiteral(ops::BundleLiteral),
        EnumLiteral(ops::EnumLiteral),
        NotU(ops::NotU),
        NotS(ops::NotS),
        NotB(ops::NotB),
        Neg(ops::Neg),
        BitAndU(ops::BitAndU),
        BitAndS(ops::BitAndS),
        BitAndB(ops::BitAndB),
        BitOrU(ops::BitOrU),
        BitOrS(ops::BitOrS),
        BitOrB(ops::BitOrB),
        BitXorU(ops::BitXorU),
        BitXorS(ops::BitXorS),
        BitXorB(ops::BitXorB),
        AddU(ops::AddU),
        AddS(ops::AddS),
        SubU(ops::SubU),
        SubS(ops::SubS),
        MulU(ops::MulU),
        MulS(ops::MulS),
        DivU(ops::DivU),
        DivS(ops::DivS),
        RemU(ops::RemU),
        RemS(ops::RemS),
        DynShlU(ops::DynShlU),
        DynShlS(ops::DynShlS),
        DynShrU(ops::DynShrU),
        DynShrS(ops::DynShrS),
        FixedShlU(ops::FixedShlU),
        FixedShlS(ops::FixedShlS),
        FixedShrU(ops::FixedShrU),
        FixedShrS(ops::FixedShrS),
        CmpLtB(ops::CmpLtB),
        CmpLeB(ops::CmpLeB),
        CmpGtB(ops::CmpGtB),
        CmpGeB(ops::CmpGeB),
        CmpEqB(ops::CmpEqB),
        CmpNeB(ops::CmpNeB),
        CmpLtU(ops::CmpLtU),
        CmpLeU(ops::CmpLeU),
        CmpGtU(ops::CmpGtU),
        CmpGeU(ops::CmpGeU),
        CmpEqU(ops::CmpEqU),
        CmpNeU(ops::CmpNeU),
        CmpLtS(ops::CmpLtS),
        CmpLeS(ops::CmpLeS),
        CmpGtS(ops::CmpGtS),
        CmpGeS(ops::CmpGeS),
        CmpEqS(ops::CmpEqS),
        CmpNeS(ops::CmpNeS),
        CastUIntToUInt(ops::CastUIntToUInt),
        CastUIntToSInt(ops::CastUIntToSInt),
        CastSIntToUInt(ops::CastSIntToUInt),
        CastSIntToSInt(ops::CastSIntToSInt),
        CastBoolToUInt(ops::CastBoolToUInt),
        CastBoolToSInt(ops::CastBoolToSInt),
        FieldAccess(ops::FieldAccess),
        VariantAccess(ops::VariantAccess),
        ArrayIndex(ops::ArrayIndex),
        DynArrayIndex(ops::DynArrayIndex),
        IntSliceU(ops::IntSliceU),
        IntSliceS(ops::IntSliceS),
    }
}

impl From<UIntValue> for ExprEnum {
    fn from(value: UIntValue) -> Self {
        ExprEnum::UIntLiteral(Intern::intern_sized(value))
    }
}

impl From<SIntValue> for ExprEnum {
    fn from(value: SIntValue) -> Self {
        ExprEnum::SIntLiteral(Intern::intern_sized(value))
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct Expr<T: Type> {
    __enum: Interned<ExprEnum>,
    __ty: T,
}

impl<T: Type> Expr<T> {
    pub fn expr_enum(this: Self) -> Interned<ExprEnum> {
        this.__enum
    }
    pub fn ty(this: Self) -> T {
        this.__ty
    }
    pub fn canonical(this: Self) -> Expr<CanonicalType> {
        Expr {
            __enum: this.__enum,
            __ty: this.__ty.canonical(),
        }
    }
    pub fn as_dyn_int(this: Self) -> Expr<T::Dyn>
    where
        T: IntType,
    {
        Expr {
            __enum: this.__enum,
            __ty: this.__ty.as_dyn_int(),
        }
    }
}

impl<T: Type, Len: Size> Expr<ArrayType<T, Len>> {
    pub fn as_dyn_array(this: Self) -> Expr<Array> {
        Expr {
            __enum: this.__enum,
            __ty: this.__ty.as_dyn_array(),
        }
    }
}

pub trait ToExpr {
    type Type: Type;
    fn to_expr(&self) -> Expr<Self::Type>;
}

impl<T: Type> ToExpr for Expr<T> {
    type Type = T;

    fn to_expr(&self) -> Expr<Self::Type> {
        *self
    }
}

impl<T: ?Sized + ToExpr> ToExpr for &'_ T {
    type Type = T::Type;

    fn to_expr(&self) -> Expr<Self::Type> {
        T::to_expr(self)
    }
}

impl<T: ?Sized + ToExpr> ToExpr for &'_ mut T {
    type Type = T::Type;

    fn to_expr(&self) -> Expr<Self::Type> {
        T::to_expr(self)
    }
}

impl<T: ?Sized + ToExpr> ToExpr for Box<T> {
    type Type = T::Type;

    fn to_expr(&self) -> Expr<Self::Type> {
        T::to_expr(self)
    }
}

impl<T: ?Sized + ToExpr + Send + Sync + 'static> ToExpr for Interned<T> {
    type Type = T::Type;

    fn to_expr(&self) -> Expr<Self::Type> {
        T::to_expr(self)
    }
}

impl<Width: Size> ToExpr for UIntValue<Width> {
    type Type = UIntType<Width>;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::UIntLiteral(self.clone().as_dyn_int().intern()).intern(),
            __ty: self.ty(),
        }
    }
}

impl<Width: Size> ToExpr for SIntValue<Width> {
    type Type = SIntType<Width>;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::SIntLiteral(self.clone().as_dyn_int().intern()).intern(),
            __ty: self.ty(),
        }
    }
}

impl ToExpr for bool {
    type Type = Bool;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::BoolLiteral(*self).intern(),
            __ty: Bool,
        }
    }
}
