// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    array::Array,
    bundle::{Bundle, BundleField, BundleType},
    enum_::{Enum, EnumType, EnumVariant},
    expr::{Expr, ExprEnum, ToExpr},
    int::{
        Bool, DynSize, IntCmp, IntType, SInt, SIntType, SIntValue, Size, UInt, UIntType, UIntValue,
    },
    intern::{Intern, Interned},
    ty::{CanonicalType, StaticType, Type},
};
use std::{
    fmt,
    ops::{
        Add, BitAnd, BitOr, BitXor, Div, Index, Mul, Neg as StdNeg, Not, Range, RangeBounds, Rem,
        Shl, Shr, Sub,
    },
};

macro_rules! forward_value_to_expr_unary_op_trait {
    (
        #[generics($($generics:tt)*)]
        #[value($Value:ty)]
        $Trait:ident::$method:ident
    ) => {
        impl<$($generics)*> $Trait for $Value
        where
            Expr<<$Value as ToExpr>::Type>: $Trait,
        {
            type Output = <Expr<<$Value as ToExpr>::Type> as $Trait>::Output;

            fn $method(self) -> Self::Output {
                $Trait::$method(self.to_expr())
            }
        }
    };
}

macro_rules! impl_unary_op_trait {
    (
        #[generics($($generics:tt)*)]
        fn $Trait:ident::$method:ident($arg:ident: $Arg:ty) -> $Output:ty {
            $($body:tt)*
        }
    ) => {
        impl<$($generics)*> $Trait for Expr<$Arg>
        {
            type Output = Expr<$Output>;

            fn $method(self) -> Self::Output {
                let $arg = self;
                $($body)*
            }
        }
    };
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct NotU<Width: Size = DynSize> {
    pub arg: Expr<UIntType<Width>>,
}

impl<Width: Size> ToExpr for NotU<Width> {
    type Type = UIntType<Width>;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::NotU(NotU {
                arg: Expr::as_dyn_int(self.arg),
            })
            .intern(),
            __ty: self.arg.__ty,
        }
    }
}

impl_unary_op_trait! {
    #[generics(Width: Size)]
    fn Not::not(arg: UIntType<Width>) -> UIntType<Width> {
        NotU { arg }.to_expr()
    }
}

forward_value_to_expr_unary_op_trait! {
    #[generics(Width: Size)]
    #[value(UIntValue<Width>)]
    Not::not
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct NotS<Width: Size = DynSize> {
    pub arg: Expr<SIntType<Width>>,
}

impl<Width: Size> ToExpr for NotS<Width> {
    type Type = UIntType<Width>;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::NotS(NotS {
                arg: Expr::as_dyn_int(self.arg),
            })
            .intern(),
            __ty: self.arg.__ty.as_same_width_uint(),
        }
    }
}

impl_unary_op_trait! {
    #[generics(Width: Size)]
    fn Not::not(arg: SIntType<Width>) -> UIntType<Width> {
        NotS { arg }.to_expr()
    }
}

forward_value_to_expr_unary_op_trait! {
    #[generics(Width: Size)]
    #[value(SIntValue<Width>)]
    Not::not
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct NotB {
    pub arg: Expr<Bool>,
}

impl ToExpr for NotB {
    type Type = Bool;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::NotB(NotB { arg: self.arg }).intern(),
            __ty: self.arg.__ty,
        }
    }
}

impl_unary_op_trait! {
    #[generics()]
    fn Not::not(arg: Bool) -> Bool {
        NotB { arg }.to_expr()
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct Neg {
    pub arg: Expr<SInt>,
}

impl ToExpr for Neg {
    type Type = SInt;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::Neg(*self).intern(),
            __ty: SInt::new_dyn(
                Expr::ty(self.arg)
                    .width()
                    .checked_add(1)
                    .expect("width too big"),
            ),
        }
    }
}

impl_unary_op_trait! {
    #[generics(Width: Size)]
    fn StdNeg::neg(arg: SIntType<Width>) -> SInt {
        Neg { arg: Expr::as_dyn_int(arg) }.to_expr()
    }
}

forward_value_to_expr_unary_op_trait! {
    #[generics(Width: Size)]
    #[value(SIntValue<Width>)]
    StdNeg::neg
}

macro_rules! impl_binary_op_trait {
    (
        #[generics($($generics:tt)*)]
        fn $Trait:ident::$method:ident($lhs:ident: $Lhs:ty, $rhs:ident: $Rhs:ty) -> $Output:ty {
            $($body:tt)*
        }
    ) => {
        impl<
                Rhs: ToExpr<Type = $Rhs>,
                $($generics)*
            > $Trait<Rhs> for Expr<$Lhs>
        {
            type Output = Expr<$Output>;

            fn $method(self, rhs: Rhs) -> Self::Output {
                let $lhs = self;
                let $rhs = rhs.to_expr();
                $($body)*
            }
        }
    };
}

macro_rules! forward_value_to_expr_binary_op_trait {
    (
        #[generics($($generics:tt)*)]
        #[lhs_value($LhsValue:ty)]
        $Trait:ident::$method:ident
    ) => {
        impl<
                Rhs,
                $($generics)*
            > $Trait<Rhs> for $LhsValue
        where
            Expr<<$LhsValue as ToExpr>::Type>: $Trait<Rhs>,
        {
            type Output = <Expr<<$LhsValue as ToExpr>::Type> as $Trait<Rhs>>::Output;

            fn $method(self, rhs: Rhs) -> Self::Output {
                $Trait::$method(self.to_expr(), rhs)
            }
        }
    };
}

macro_rules! binary_op_bitwise {
    ($name:ident, $ty:ident, bool, $Trait:ident::$method:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty,
                }
            }
        }

        /// intentionally only implemented for expressions, not rust's bool type,
        /// since that helps avoid using `==`/`!=` in hdl boolean expressions, which doesn't do
        /// what is usually wanted.
        impl $Trait for Expr<Bool> {
            type Output = Expr<Bool>;

            fn $method(self, rhs: Expr<Bool>) -> Expr<Bool> {
                $name { lhs: self, rhs }.to_expr()
            }
        }
    };
    ($name:ident, $ty:ident, $value:ident, $Trait:ident::$method:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = UInt;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: UInt::new_dyn(Expr::ty(self.lhs).width().max(Expr::ty(self.lhs).width())),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: Size, RhsWidth: Size)]
            fn $Trait::$method(lhs: $ty<LhsWidth>, rhs: $ty<RhsWidth>) -> UInt {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }

        forward_value_to_expr_binary_op_trait! {
            #[generics(LhsWidth: Size)]
            #[lhs_value($value<LhsWidth>)]
            $Trait::$method
        }
    };
}

binary_op_bitwise!(BitAndU, UIntType, UIntValue, BitAnd::bitand);
binary_op_bitwise!(BitAndS, SIntType, SIntValue, BitAnd::bitand);
binary_op_bitwise!(BitAndB, Bool, bool, BitAnd::bitand);
binary_op_bitwise!(BitOrU, UIntType, UIntValue, BitOr::bitor);
binary_op_bitwise!(BitOrS, SIntType, SIntValue, BitOr::bitor);
binary_op_bitwise!(BitOrB, Bool, bool, BitOr::bitor);
binary_op_bitwise!(BitXorU, UIntType, UIntValue, BitXor::bitxor);
binary_op_bitwise!(BitXorS, SIntType, SIntValue, BitXor::bitxor);
binary_op_bitwise!(BitXorB, Bool, bool, BitXor::bitxor);

macro_rules! binary_op_add_sub {
    ($name:ident, $ty:ident, $value:ident, $Trait:ident::$method:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty::new_dyn(
                        Expr::ty(self.lhs)
                            .width()
                            .max(Expr::ty(self.rhs).width())
                            .checked_add(1)
                            .expect("width too big"),
                    ),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: Size, RhsWidth: Size)]
            fn $Trait::$method(lhs: $ty<LhsWidth>, rhs: $ty<RhsWidth>) -> $ty {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }

        forward_value_to_expr_binary_op_trait! {
            #[generics(LhsWidth: Size)]
            #[lhs_value($value<LhsWidth>)]
            $Trait::$method
        }
    };
}

binary_op_add_sub!(AddU, UIntType, UIntValue, Add::add);
binary_op_add_sub!(AddS, SIntType, SIntValue, Add::add);
binary_op_add_sub!(SubU, UIntType, UIntValue, Sub::sub);
binary_op_add_sub!(SubS, SIntType, SIntValue, Sub::sub);

macro_rules! binary_op_mul {
    ($name:ident, $ty:ident, $value:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty::new_dyn(
                        Expr::ty(self.lhs)
                            .width()
                            .checked_add(Expr::ty(self.rhs).width())
                            .expect("width too big"),
                    ),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: Size, RhsWidth: Size)]
            fn Mul::mul(lhs: $ty<LhsWidth>, rhs: $ty<RhsWidth>) -> $ty {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }

        forward_value_to_expr_binary_op_trait! {
            #[generics(LhsWidth: Size)]
            #[lhs_value($value<LhsWidth>)]
            Mul::mul
        }
    };
}

binary_op_mul!(MulU, UIntType, UIntValue);
binary_op_mul!(MulS, SIntType, SIntValue);

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct DivU<LhsWidth: Size = DynSize> {
    pub lhs: Expr<UIntType<LhsWidth>>,
    pub rhs: Expr<UInt>,
}

impl<LhsWidth: Size> ToExpr for DivU<LhsWidth> {
    type Type = UIntType<LhsWidth>;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::DivU(DivU {
                lhs: Expr::as_dyn_int(self.lhs),
                rhs: self.rhs,
            })
            .intern(),
            __ty: Expr::ty(self.lhs),
        }
    }
}

impl_binary_op_trait! {
    #[generics(LhsWidth: Size, RhsWidth: Size)]
    fn Div::div(lhs: UIntType<LhsWidth>, rhs: UIntType<RhsWidth>) -> UIntType<LhsWidth> {
        DivU {
            lhs,
            rhs: Expr::as_dyn_int(rhs),
        }
        .to_expr()
    }
}

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: Size)]
    #[lhs_value(UIntValue<LhsWidth>)]
    Div::div
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct DivS {
    pub lhs: Expr<SInt>,
    pub rhs: Expr<SInt>,
}

impl ToExpr for DivS {
    type Type = SInt;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::DivS(*self).intern(),
            __ty: SInt::new_dyn(
                Expr::ty(self.lhs)
                    .width()
                    .checked_add(1)
                    .expect("width too big"),
            ),
        }
    }
}

impl_binary_op_trait! {
    #[generics(LhsWidth: Size, RhsWidth: Size)]
    fn Div::div(lhs: SIntType<LhsWidth>, rhs: SIntType<RhsWidth>) -> SInt {
        DivS {
            lhs: Expr::as_dyn_int(lhs),
            rhs: Expr::as_dyn_int(rhs),
        }
        .to_expr()
    }
}

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: Size)]
    #[lhs_value(SIntValue<LhsWidth>)]
    Div::div
}

macro_rules! binary_op_rem {
    ($name:ident, $ty:ident, $value:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty::new_dyn(Expr::ty(self.lhs).width().min(Expr::ty(self.rhs).width())),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: Size, RhsWidth: Size)]
            fn Rem::rem(lhs: $ty<LhsWidth>, rhs: $ty<RhsWidth>) -> $ty {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }

        forward_value_to_expr_binary_op_trait! {
            #[generics(LhsWidth: Size)]
            #[lhs_value($value<LhsWidth>)]
            Rem::rem
        }
    };
}

binary_op_rem!(RemU, UIntType, UIntValue);
binary_op_rem!(RemS, SIntType, SIntValue);

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct BundleLiteral<T: BundleType = Bundle> {
    ty: T,
    field_values: Interned<[Expr<CanonicalType>]>,
}

impl<T: BundleType> BundleLiteral<T> {
    #[track_caller]
    pub fn new(ty: T, field_values: &[Expr<CanonicalType>]) -> Self {
        let fields = ty.fields();
        assert_eq!(fields.len(), field_values.len());
        for (&field, &field_value) in fields.iter().zip(field_values) {
            assert_eq!(
                field.ty,
                Expr::ty(field_value),
                "field's type doesn't match value's type: field name {:?}",
                field.name
            );
        }
        Self {
            ty,
            field_values: field_values.intern(),
        }
    }
    pub fn ty(self) -> T {
        self.ty
    }
    pub fn field_values(self) -> Interned<[Expr<CanonicalType>]> {
        self.field_values
    }
}

impl<T: BundleType> ToExpr for BundleLiteral<T> {
    type Type = T;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::BundleLiteral(BundleLiteral {
                ty: Bundle::new(self.ty.fields()),
                field_values: self.field_values,
            })
            .intern(),
            __ty: self.ty,
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct EnumLiteral<T: EnumType = Enum> {
    ty: T,
    variant_index: usize,
    variant_value: Option<Expr<CanonicalType>>,
}

impl<T: EnumType> EnumLiteral<T> {
    #[track_caller]
    pub fn new(ty: T, variant_index: usize, variant_value: Option<Expr<CanonicalType>>) -> Self {
        let variants = ty.variants();
        assert!(variant_index < variants.len());
        assert_eq!(
            variants[variant_index].ty,
            variant_value.map(Expr::ty),
            "variant's type doesn't match value's type: variant name {:?}",
            variants[variant_index].name
        );
        Self {
            ty,
            variant_index,
            variant_value,
        }
    }
    pub fn ty(self) -> T {
        self.ty
    }
    pub fn variant_index(self) -> usize {
        self.variant_index
    }
    pub fn variant_value(self) -> Option<Expr<CanonicalType>> {
        self.variant_value
    }
}

impl<T: EnumType> ToExpr for EnumLiteral<T> {
    type Type = T;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::EnumLiteral(EnumLiteral {
                ty: Enum::new(self.ty.variants()),
                variant_index: self.variant_index,
                variant_value: self.variant_value,
            })
            .intern(),
            __ty: self.ty,
        }
    }
}

macro_rules! impl_dyn_shl {
    ($name:ident, $ty:ident, $value:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<UInt>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            #[track_caller]
            fn to_expr(&self) -> Expr<Self::Type> {
                let Some(pow2_rhs_width) = Expr::ty(self.rhs)
                    .width()
                    .try_into()
                    .ok()
                    .and_then(|v| 2usize.checked_pow(v))
                else {
                    panic!(
                        "dynamic left-shift amount's bit-width is too big, try casting the shift \
                        amount to a smaller bit-width before shifting"
                    );
                };
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty::new_dyn(
                        (pow2_rhs_width - 1)
                            .checked_add(Expr::ty(self.lhs).width())
                            .expect("width too big"),
                    ),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: Size, RhsWidth: Size)]
            fn Shl::shl(lhs: $ty<LhsWidth>, rhs: UIntType<RhsWidth>) -> $ty {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }
    };
}

impl_dyn_shl!(DynShlU, UIntType, UIntValue);
impl_dyn_shl!(DynShlS, SIntType, SIntValue);

macro_rules! impl_dyn_shr {
    ($name:ident, $ty:ident, $value:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name<LhsWidth: Size = DynSize> {
            pub lhs: Expr<$ty<LhsWidth>>,
            pub rhs: Expr<UInt>,
        }

        impl<LhsWidth: Size> ToExpr for $name<LhsWidth> {
            type Type = $ty<LhsWidth>;

            #[track_caller]
            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name($name {
                        lhs: Expr::as_dyn_int(self.lhs),
                        rhs: self.rhs,
                    })
                    .intern(),
                    __ty: Expr::ty(self.lhs),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: Size, RhsWidth: Size)]
            fn Shr::shr(lhs: $ty<LhsWidth>, rhs: UIntType<RhsWidth>) -> $ty<LhsWidth> {
                $name {
                    lhs,
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }
    };
}

impl_dyn_shr!(DynShrU, UIntType, UIntValue);
impl_dyn_shr!(DynShrS, SIntType, SIntValue);

fn fixed_shr_width(lhs_width: usize, rhs: usize) -> Option<usize> {
    Some(lhs_width.saturating_sub(rhs).max(1))
}

macro_rules! binary_op_fixed_shift {
    ($name:ident, $ty:ident, $value:ident, $width_fn:path, $Trait:ident::$method:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: usize,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty::new_dyn(
                        $width_fn(Expr::ty(self.lhs).width(), self.rhs).expect("width too big"),
                    ),
                }
            }
        }

        impl<LhsWidth: Size> $Trait<usize> for Expr<$ty<LhsWidth>> {
            type Output = Expr<$ty>;

            fn $method(self, rhs: usize) -> Self::Output {
                $name {
                    lhs: Expr::as_dyn_int(self),
                    rhs,
                }
                .to_expr()
            }
        }
    };
}

binary_op_fixed_shift!(FixedShlU, UIntType, UIntValue, usize::checked_add, Shl::shl);
binary_op_fixed_shift!(FixedShlS, SIntType, SIntValue, usize::checked_add, Shl::shl);
binary_op_fixed_shift!(FixedShrU, UIntType, UIntValue, fixed_shr_width, Shr::shr);
binary_op_fixed_shift!(FixedShrS, SIntType, SIntValue, fixed_shr_width, Shr::shr);

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: Size)]
    #[lhs_value(UIntValue<LhsWidth>)]
    Shl::shl
}

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: Size)]
    #[lhs_value(SIntValue<LhsWidth>)]
    Shl::shl
}

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: Size)]
    #[lhs_value(UIntValue<LhsWidth>)]
    Shr::shr
}

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: Size)]
    #[lhs_value(SIntValue<LhsWidth>)]
    Shr::shr
}

pub trait IntCmpExpr<Rhs: Type>: Type {
    fn cmp_eq(lhs: Expr<Self>, rhs: Expr<Rhs>) -> Expr<Bool>;
    fn cmp_ne(lhs: Expr<Self>, rhs: Expr<Rhs>) -> Expr<Bool>;
    fn cmp_lt(lhs: Expr<Self>, rhs: Expr<Rhs>) -> Expr<Bool>;
    fn cmp_le(lhs: Expr<Self>, rhs: Expr<Rhs>) -> Expr<Bool>;
    fn cmp_gt(lhs: Expr<Self>, rhs: Expr<Rhs>) -> Expr<Bool>;
    fn cmp_ge(lhs: Expr<Self>, rhs: Expr<Rhs>) -> Expr<Bool>;
}

impl<Lhs: ToExpr, Rhs: ToExpr> IntCmp<Rhs> for Lhs
where
    Lhs::Type: IntCmpExpr<Rhs::Type>,
{
    fn cmp_eq(self, rhs: Rhs) -> Expr<Bool> {
        IntCmpExpr::cmp_eq(self.to_expr(), rhs.to_expr())
    }
    fn cmp_ne(self, rhs: Rhs) -> Expr<Bool> {
        IntCmpExpr::cmp_ne(self.to_expr(), rhs.to_expr())
    }
    fn cmp_lt(self, rhs: Rhs) -> Expr<Bool> {
        IntCmpExpr::cmp_lt(self.to_expr(), rhs.to_expr())
    }
    fn cmp_le(self, rhs: Rhs) -> Expr<Bool> {
        IntCmpExpr::cmp_le(self.to_expr(), rhs.to_expr())
    }
    fn cmp_gt(self, rhs: Rhs) -> Expr<Bool> {
        IntCmpExpr::cmp_gt(self.to_expr(), rhs.to_expr())
    }
    fn cmp_ge(self, rhs: Rhs) -> Expr<Bool> {
        IntCmpExpr::cmp_ge(self.to_expr(), rhs.to_expr())
    }
}

macro_rules! impl_compare_op {
    (
        $(#[width($LhsWidth:ident, $RhsWidth:ident)])?
        #[dyn_type($DynTy:ident)]
        #[to_dyn_type($lhs:ident => $dyn_lhs:expr, $rhs:ident => $dyn_rhs:expr)]
        #[type($Lhs:ty, $Rhs:ty)]
        $(
            struct $name:ident;
            fn $method:ident();
        )*
    ) => {
        $(#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$DynTy>,
            pub rhs: Expr<$DynTy>,
        }

        impl ToExpr for $name {
            type Type = Bool;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: Bool,
                }
            }
        })*

        impl$(<$LhsWidth: Size, $RhsWidth: Size>)? IntCmpExpr<$Rhs> for $Lhs {
            $(fn $method($lhs: Expr<Self>, $rhs: Expr<$Rhs>) -> Expr<Bool> {
                $name {
                    lhs: $dyn_lhs,
                    rhs: $dyn_rhs,
                }.to_expr()
            })*
        }
    };
}

impl_compare_op! {
    #[dyn_type(Bool)]
    #[to_dyn_type(lhs => lhs, rhs => rhs)]
    #[type(Bool, Bool)]
    struct CmpEqB; fn cmp_eq();
    struct CmpNeB; fn cmp_ne();
    struct CmpLtB; fn cmp_lt();
    struct CmpLeB; fn cmp_le();
    struct CmpGtB; fn cmp_gt();
    struct CmpGeB; fn cmp_ge();
}

impl_compare_op! {
    #[width(LhsWidth, RhsWidth)]
    #[dyn_type(UInt)]
    #[to_dyn_type(lhs => Expr::as_dyn_int(lhs), rhs => Expr::as_dyn_int(rhs))]
    #[type(UIntType<LhsWidth>, UIntType<RhsWidth>)]
    struct CmpEqU; fn cmp_eq();
    struct CmpNeU; fn cmp_ne();
    struct CmpLtU; fn cmp_lt();
    struct CmpLeU; fn cmp_le();
    struct CmpGtU; fn cmp_gt();
    struct CmpGeU; fn cmp_ge();
}

impl_compare_op! {
    #[width(LhsWidth, RhsWidth)]
    #[dyn_type(SInt)]
    #[to_dyn_type(lhs => Expr::as_dyn_int(lhs), rhs => Expr::as_dyn_int(rhs))]
    #[type(SIntType<LhsWidth>, SIntType<RhsWidth>)]
    struct CmpEqS; fn cmp_eq();
    struct CmpNeS; fn cmp_ne();
    struct CmpLtS; fn cmp_lt();
    struct CmpLeS; fn cmp_le();
    struct CmpGtS; fn cmp_gt();
    struct CmpGeS; fn cmp_ge();
}

pub trait ExprCastTo<ToType: Type>: Type {
    fn cast_to(src: Expr<Self>, to_type: ToType) -> Expr<ToType>;
}

pub trait CastTo: Sized {
    type FromType: Type;
    fn cast_to<ToType: Type>(self, to_type: ToType) -> Expr<ToType>
    where
        Self::FromType: ExprCastTo<ToType>;
    fn cast_to_static<ToType: StaticType>(self) -> Expr<ToType>
    where
        Self::FromType: ExprCastTo<ToType>;
}

impl ExprCastTo<Bool> for Bool {
    fn cast_to(src: Expr<Self>, _to_type: Bool) -> Expr<Bool> {
        src
    }
}

impl<FromType: Type> CastTo for Expr<FromType> {
    type FromType = FromType;
    fn cast_to<ToType: Type>(self, to_type: ToType) -> Expr<ToType>
    where
        FromType: ExprCastTo<ToType>,
    {
        ExprCastTo::cast_to(self, to_type)
    }
    fn cast_to_static<ToType: StaticType>(self) -> Expr<ToType>
    where
        FromType: ExprCastTo<ToType>,
    {
        self.cast_to(ToType::static_type())
    }
}

macro_rules! impl_cast_int_op {
    ($name:ident, $from:ident, $to:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name<ToWidth: Size = DynSize> {
            pub arg: Expr<$from>,
            pub ty: $to<ToWidth>,
        }

        impl<ToWidth: Size> ToExpr for $name<ToWidth> {
            type Type = $to<ToWidth>;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name($name {
                        arg: self.arg,
                        ty: self.ty.as_dyn_int(),
                    })
                    .intern(),
                    __ty: self.ty,
                }
            }
        }

        impl<FromWidth: Size, ToWidth: Size> ExprCastTo<$to<ToWidth>> for $from<FromWidth> {
            fn cast_to(src: Expr<Self>, to_type: $to<ToWidth>) -> Expr<$to<ToWidth>> {
                $name {
                    arg: Expr::as_dyn_int(src),
                    ty: to_type,
                }
                .to_expr()
            }
        }
    };
}

impl_cast_int_op!(CastUIntToUInt, UIntType, UIntType);
impl_cast_int_op!(CastUIntToSInt, UIntType, SIntType);
impl_cast_int_op!(CastSIntToUInt, SIntType, UIntType);
impl_cast_int_op!(CastSIntToSInt, SIntType, SIntType);

macro_rules! impl_cast_bit_op {
    ($name:ident, $from:ty, $to:ty) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub arg: Expr<$from>,
            pub ty: $to,
        }

        impl ToExpr for $name {
            type Type = $to;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name($name {
                        arg: self.arg,
                        ty: self.ty.as_dyn_int(),
                    })
                    .intern(),
                    __ty: self.ty,
                }
            }
        }

        impl ExprCastTo<$to> for $from {
            fn cast_to(src: Expr<Self>, to_type: $to) -> Expr<$to> {
                $name {
                    arg: src,
                    ty: to_type,
                }
                .to_expr()
            }
        }
    };
}

impl_cast_bit_op!(CastBoolToUInt, Bool, UInt);
impl_cast_bit_op!(CastBoolToSInt, Bool, SInt);

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct FieldAccess<FieldType: Type = CanonicalType> {
    base: Expr<Bundle>,
    field_index: usize,
    field: BundleField,
    field_type: FieldType,
}

impl<FieldType: Type> FieldAccess<FieldType> {
    #[track_caller]
    pub fn new_by_index(base: Expr<Bundle>, field_index: usize) -> Self {
        let field = Expr::ty(base).fields()[field_index];
        let field_type = FieldType::from_canonical(field.ty);
        Self {
            base,
            field_index,
            field,
            field_type,
        }
    }
    #[track_caller]
    pub fn new_by_name(base: Expr<Bundle>, name: Interned<str>) -> Self {
        let base_ty = Expr::ty(base);
        let Some(field_index) = base_ty.name_indexes().get(&name) else {
            panic!("unknown field {name:?}: in {base_ty:?}");
        };
        Self::new_by_index(base, *field_index)
    }
    pub fn base(self) -> Expr<Bundle> {
        self.base
    }
    pub fn field_index(self) -> usize {
        self.field_index
    }
    pub fn field_name(self) -> Interned<str> {
        self.field.name
    }
    pub fn field(self) -> BundleField {
        self.field
    }
    pub fn field_type(self) -> FieldType {
        self.field_type
    }
}

impl<FieldType: Type> ToExpr for FieldAccess<FieldType> {
    type Type = FieldType;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::FieldAccess(FieldAccess {
                base: self.base,
                field_index: self.field_index,
                field: self.field,
                field_type: self.field_type.canonical(),
            })
            .intern(),
            __ty: self.field_type,
        }
    }
}

impl<T: BundleType> Expr<T> {
    pub fn as_bundle(this: Self) -> Expr<Bundle> {
        Expr {
            __enum: this.__enum,
            __ty: Bundle::new(this.__ty.fields()),
        }
    }
    #[track_caller]
    pub fn field<FieldType: Type>(this: Self, name: &str) -> Expr<FieldType> {
        FieldAccess::new_by_name(Self::as_bundle(this), name.intern()).to_expr()
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct VariantAccess<VariantType: Type = CanonicalType> {
    base: Expr<Enum>,
    variant_index: usize,
    variant: EnumVariant,
    variant_type: Option<VariantType>,
}

impl<VariantType: Type> VariantAccess<VariantType> {
    #[track_caller]
    pub fn new_unchecked(base: Expr<Enum>, variant_index: usize) -> Self {
        let variant = Expr::ty(base).variants()[variant_index];
        let variant_type = variant.ty.map(VariantType::from_canonical);
        Self {
            base,
            variant_index,
            variant,
            variant_type,
        }
    }
    pub fn base(self) -> Expr<Enum> {
        self.base
    }
    pub fn variant_index(self) -> usize {
        self.variant_index
    }
    pub fn variant_name(self) -> Interned<str> {
        self.variant.name
    }
    pub fn variant(self) -> EnumVariant {
        self.variant
    }
    pub fn variant_type(self) -> Option<VariantType> {
        self.variant_type
    }
}

impl<VariantType: Type> ToExpr for VariantAccess<VariantType> {
    type Type = VariantType;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::VariantAccess(VariantAccess {
                base: self.base,
                variant_index: self.variant_index,
                variant: self.variant,
                variant_type: self.variant.ty,
            })
            .intern(),
            __ty: self
                .variant_type
                .unwrap_or_else(|| VariantType::from_canonical(().canonical())),
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct ArrayIndex<ElementType: Type = CanonicalType> {
    base: Expr<Array>,
    element_index: usize,
    element_type: ElementType,
}

impl<ElementType: Type> ArrayIndex<ElementType> {
    #[track_caller]
    pub fn new_unchecked(base: Expr<Array>, element_index: usize) -> Self {
        let base_ty = Expr::ty(base);
        assert!(element_index < base_ty.len());
        let element_type = ElementType::from_canonical(*base_ty.element());
        Self {
            base,
            element_index,
            element_type,
        }
    }
    pub fn base(self) -> Expr<Array> {
        self.base
    }
    pub fn element_index(self) -> usize {
        self.element_index
    }
    pub fn element_type(self) -> ElementType {
        self.element_type
    }
}

impl<ElementType: Type> ToExpr for ArrayIndex<ElementType> {
    type Type = ElementType;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::ArrayIndex(ArrayIndex {
                base: self.base,
                element_index: self.element_index,
                element_type: *Expr::ty(self.base).element(),
            })
            .intern(),
            __ty: self.element_type,
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct DynArrayIndex<ElementType: Type = CanonicalType> {
    base: Expr<Array>,
    element_index: Expr<UInt>,
    element_type: ElementType,
}

impl<ElementType: Type> DynArrayIndex<ElementType> {
    #[track_caller]
    pub fn new_unchecked(base: Expr<Array>, element_index: Expr<UInt>) -> Self {
        let base_ty = Expr::ty(base);
        let element_type = ElementType::from_canonical(*base_ty.element());
        Self {
            base,
            element_index,
            element_type,
        }
    }
    pub fn base(self) -> Expr<Array> {
        self.base
    }
    pub fn element_index(self) -> Expr<UInt> {
        self.element_index
    }
    pub fn element_type(self) -> ElementType {
        self.element_type
    }
}

impl<ElementType: Type> ToExpr for DynArrayIndex<ElementType> {
    type Type = ElementType;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::DynArrayIndex(DynArrayIndex {
                base: self.base,
                element_index: self.element_index,
                element_type: *Expr::ty(self.base).element(),
            })
            .intern(),
            __ty: self.element_type,
        }
    }
}

pub trait ExprIndex<I>: Type {
    type Output: Type;
    fn expr_index(this: &Expr<Self>, index: I) -> &Expr<Self::Output>;
}

impl<T: ExprIndex<I>, I> Index<I> for Expr<T> {
    type Output = Expr<T::Output>;

    #[track_caller]
    fn index(&self, index: I) -> &Self::Output {
        T::expr_index(self, index)
    }
}

mod sealed {
    pub trait BuiltInRangeSealed {}
}

pub trait BuiltInRange<T: ?Sized>: RangeBounds<T> + sealed::BuiltInRangeSealed {}

impl<I: ?Sized, T: RangeBounds<I> + ?Sized + sealed::BuiltInRangeSealed> BuiltInRange<I> for T {}

macro_rules! impl_built_in_range {
    (<$T:ident> $ty:ty) => {
        impl<$T> sealed::BuiltInRangeSealed for $ty {}
    };
    ($ty:ty) => {
        impl sealed::BuiltInRangeSealed for $ty {}
    };
}

impl_built_in_range!(<T> (std::ops::Bound<T>, std::ops::Bound<T>));
impl_built_in_range!(<T> std::ops::Range<T>);
impl_built_in_range!(<T> std::ops::RangeFrom<T>);
impl_built_in_range!(std::ops::RangeFull);
impl_built_in_range!(<T> std::ops::RangeInclusive<T>);
impl_built_in_range!(<T> std::ops::RangeTo<T>);
impl_built_in_range!(<T> std::ops::RangeToInclusive<T>);

macro_rules! impl_int_slice {
    ($name:ident, $ty:ident, $value:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash)]
        pub struct $name {
            base: Expr<$ty>,
            range_start: usize,
            range_end: usize,
        }

        impl fmt::Debug for $name {
            fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
                let Self {
                    base,
                    range_start: _,
                    range_end: _,
                } = self;
                f.debug_struct(stringify!($name))
                    .field("base", base)
                    .field("range", &self.range())
                    .finish()
            }
        }

        impl $name {
            pub fn new(base: Expr<$ty>, range: Range<usize>) -> Self {
                assert!(
                    range.start <= range.end && range.end <= Expr::ty(base).width(),
                    "invalid range"
                );
                Self {
                    base,
                    range_start: range.start,
                    range_end: range.end,
                }
            }
            pub fn base(self) -> Expr<$ty> {
                self.base
            }
            pub fn range(self) -> Range<usize> {
                self.range_start..self.range_end
            }
        }

        impl ToExpr for $name {
            type Type = UInt;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: UInt::new_dyn(self.range().len()),
                }
            }
        }

        impl<Width: Size, I: BuiltInRange<usize>> ExprIndex<I> for $ty<Width> {
            type Output = UInt;

            #[track_caller]
            fn expr_index(this: &Expr<Self>, index: I) -> &Expr<Self::Output> {
                let base = Expr::as_dyn_int(*this);
                let base_ty = Expr::ty(base);
                let range = base_ty.slice_index_to_range(index);
                Interned::<_>::into_inner($name::new(base, range).to_expr().intern_sized())
            }
        }
    };
}

impl_int_slice!(IntSliceU, UIntType, UIntValue);
impl_int_slice!(IntSliceS, SIntType, SIntValue);
