// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    array::Array,
    bundle::Bundle,
    enum_::Enum,
    expr::Expr,
    int::{Bool, SInt, UInt},
    module::{ModuleBuilder, NormalModule},
    source_location::SourceLocation,
};
use std::{fmt, hash::Hash, iter::FusedIterator};

#[derive(Copy, Clone, Hash, PartialEq, Eq, Debug)]
#[non_exhaustive]
pub struct TypeProperties {
    pub is_passive: bool,
    pub is_storable: bool,
    pub is_castable_from_bits: bool,
    pub bit_width: usize,
}

#[derive(Copy, Clone, Hash, PartialEq, Eq, Debug)]
pub enum CanonicalType {
    UInt(UInt),
    SInt(SInt),
    Bool(Bool),
    Array(Array),
    Enum(Enum),
    Bundle(Bundle),
}

impl CanonicalType {
    pub fn type_properties(self) -> TypeProperties {
        match self {
            CanonicalType::UInt(v) => v.type_properties(),
            CanonicalType::SInt(v) => v.type_properties(),
            CanonicalType::Bool(v) => v.type_properties(),
            CanonicalType::Array(v) => v.type_properties(),
            CanonicalType::Enum(v) => v.type_properties(),
            CanonicalType::Bundle(v) => v.type_properties(),
        }
    }
    pub fn is_passive(self) -> bool {
        self.type_properties().is_passive
    }
    pub fn is_storable(self) -> bool {
        self.type_properties().is_storable
    }
    pub fn is_castable_from_bits(self) -> bool {
        self.type_properties().is_castable_from_bits
    }
    pub fn bit_width(self) -> usize {
        self.type_properties().bit_width
    }
}

pub trait MatchVariantAndInactiveScope: Sized {
    type MatchVariant: 'static + Send + Sync;
    type MatchActiveScope;
    fn match_activate_scope(self) -> (Self::MatchVariant, Self::MatchActiveScope);
}

#[derive(Debug)]
pub struct MatchVariantWithoutScope<T: 'static + Send + Sync>(pub T);

impl<T: 'static + Send + Sync> MatchVariantAndInactiveScope for MatchVariantWithoutScope<T> {
    type MatchVariant = T;
    type MatchActiveScope = ();

    fn match_activate_scope(self) -> (Self::MatchVariant, Self::MatchActiveScope) {
        (self.0, ())
    }
}

pub trait Type: Copy + Hash + Eq + fmt::Debug + Send + Sync + 'static {
    type MaskType: Type<MaskType = Self::MaskType>;
    type MatchVariant: 'static + Send + Sync;
    type MatchActiveScope;
    type MatchVariantAndInactiveScope: MatchVariantAndInactiveScope<
        MatchVariant = Self::MatchVariant,
        MatchActiveScope = Self::MatchActiveScope,
    >;
    type MatchVariantsIter: Iterator<Item = Self::MatchVariantAndInactiveScope>
        + ExactSizeIterator
        + FusedIterator
        + DoubleEndedIterator;
    fn match_variants(
        this: Expr<Self>,
        module_builder: &mut ModuleBuilder<NormalModule>,
        source_location: SourceLocation,
    ) -> Self::MatchVariantsIter;
    fn mask_type(&self) -> Self::MaskType;
    fn canonical(&self) -> CanonicalType;
    fn from_canonical(canonical_type: CanonicalType) -> Self;
    fn source_location() -> SourceLocation;
}

macro_rules! impl_match_variant_as_self {
    () => {
        type MatchVariant = crate::expr::Expr<Self>;
        type MatchActiveScope = ();
        type MatchVariantAndInactiveScope = crate::ty::MatchVariantWithoutScope<Self::MatchVariant>;
        type MatchVariantsIter = std::iter::Once<Self::MatchVariantAndInactiveScope>;
        fn match_variants(
            this: Expr<Self>,
            module_builder: &mut crate::module::ModuleBuilder<crate::module::NormalModule>,
            source_location: crate::source_location::SourceLocation,
        ) -> Self::MatchVariantsIter {
            let _ = module_builder;
            let _ = source_location;
            std::iter::once(crate::ty::MatchVariantWithoutScope(this))
        }
    };
}

pub(crate) use impl_match_variant_as_self;

pub trait TypeWithDeref: Type {
    fn expr_deref(this: &Expr<Self>) -> &Self::MatchVariant;
}

impl Type for CanonicalType {
    type MaskType = CanonicalType;
    impl_match_variant_as_self!();
    fn mask_type(&self) -> Self::MaskType {
        match self {
            CanonicalType::UInt(v) => v.mask_type().canonical(),
            CanonicalType::SInt(v) => v.mask_type().canonical(),
            CanonicalType::Bool(v) => v.mask_type().canonical(),
            CanonicalType::Array(v) => v.mask_type().canonical(),
            CanonicalType::Enum(v) => v.mask_type().canonical(),
            CanonicalType::Bundle(v) => v.mask_type().canonical(),
        }
    }
    fn canonical(&self) -> CanonicalType {
        *self
    }
    fn from_canonical(canonical_type: CanonicalType) -> Self {
        canonical_type
    }
    fn source_location() -> SourceLocation {
        SourceLocation::builtin()
    }
}

pub trait StaticType: Type {
    fn static_type() -> Self;
}
